/* Copyright (C) 2001 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "samhain.xpm"
#include "samhain_warn.xpm"
#include "samhain_monitor.h"
#include <gnome.h>
#include <applet-widget.h>


#define GDIE(arg)  \
do { \
   g_warning("%s (%s, %d): %s: %s\n", \
           PACKAGE, __FILE__, __LINE__, arg, g_strerror(errno)); \
   exit(EXIT_FAILURE); \
} while (1 == 0)


static int the_child_pid; 

/* Signal handler to handle the SIGCHLD signal
 * 'catch_child' is the signal number, reuse as dummy variable in waitpid()
 */
void catch_child(int child_status)
{
  waitpid((-1), &child_status, WNOHANG|WUNTRACED);
  return;
}

/* do not re-enter the signal handler
 */
volatile int my_sig_block = 0;


/* Signal handler to kill the child
 */
void kill_child(int mysignal)
{
  /* kill the child, the reset the signal handler
   * to default and raise the signal
   */
  struct sigaction act;
  if (my_sig_block == 1)
    return;
  else
    my_sig_block = 1;
  act.sa_handler = SIG_DFL;
  kill(the_child_pid, SIGTERM);
  sigaction (mysignal, &act, NULL);
  raise(mysignal);
  my_sig_block = 0;
  return;
}

/* Called from main - prepare to handle the SIGCHLD signal
 */
void init_sigchld ()
{
  struct sigaction act;

  act.sa_handler = catch_child;
  sigemptyset( &act.sa_mask );
  act.sa_flags = 0;

  sigaction(SIGCHLD, &act, NULL);

  act.sa_handler = kill_child;
  sigemptyset( &act.sa_mask );
  act.sa_flags = 0;

  sigaction(SIGHUP,  &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGABRT, &act, NULL);

  sigaction(SIGFPE,  &act, NULL);
  sigaction(SIGILL,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);

  return;
}


/**********************************************************************
 *
 *  CHILD PROCESS - EXECUTED WITH SUPERUSER PRIVILEGES
 *
 *
 */

/* The identifier of the message queue
 */
static int msgid = -1;

/* Open the SysV message queue, creating it when neccesary
 */
int open_ipc(void)
{
  key_t            key;
  int              retval = 0;

  /* get key
   */
  key = ftok ("/tmp", '#');
  if (key == (key_t) -1)
    {
      perror("ftok");
      _exit(EXIT_FAILURE);
    }


  /* get message identifier, check whether queue exists already
   */
  msgid = msgget (key, IPC_CREAT|MSGMODE|IPC_EXCL);
  if (msgid < 0 && errno == EEXIST)
    {
      /* the queue exists, return 1 and open
       */
      retval = 1;
      msgid = msgget (key, IPC_CREAT|MSGMODE);
    }

  if (msgid < 0)
    {
      fprintf(stderr, "msgget: uid: %d  euid: %d\n", getuid(), geteuid()); 
      perror("msgget");
      _exit(EXIT_FAILURE);
    }

  return retval;
}

/* Close the SysV message queue
 */
void close_ipc (void)
{
  if (msgid != (-1))
    (void) msgctl (msgid, IPC_RMID, NULL);
  return;
}

/* Signal handler to remove the message queue
 */
void ipc_sighandler (int mysignal)
{
  /* remove the message queue, the reset the signal handler
   * to default and raise the signal
   */
  struct sigaction act;
  if (my_sig_block == 1)
    return;
  else
    my_sig_block = 1;
  act.sa_handler = SIG_DFL;
  close_ipc();
  sigaction (mysignal, &act, NULL);
  raise(mysignal);
  my_sig_block = 0;
}

/* Called from child - set up signal handling for cleanup on exit
 */
void init_child ()
{
  struct sigaction act;

  act.sa_handler = ipc_sighandler;
  sigemptyset( &act.sa_mask );
  act.sa_flags = 0;

  sigaction(SIGHUP,  &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGABRT, &act, NULL);

  sigaction(SIGFPE,  &act, NULL);
  sigaction(SIGILL,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);

  /* we cannot expect the child to die from SIGPIPE, because if no-one
   * writes to the message queue, the process may sit there forever
   */
  sigaction(SIGPIPE, &act, NULL);

  return;
}


/*  read from the SysV message queue, and pipe all messages
 *  into the main program.
 */
int msg_handle(int * pipes)
{
  struct msgbuf   *recv_msg = NULL;
  char            *outbuf   = NULL;
  int              rc       = -1;
  int              status;

  init_child();

  /* struct msgbuf {
   *   long mtype;
   *   char mtext[1];  <-- sizeof(mtext) will be  1+MY_MAX_MSG
   * }
   */
  recv_msg = (struct msgbuf*) malloc (sizeof(struct msgbuf) + MY_MAX_MSG);
  if (NULL == recv_msg)
    {
      perror("malloc");
      _exit(EXIT_FAILURE);
    }

  memset (recv_msg->mtext, '\0', MY_MAX_MSG+1);

  outbuf = (char*) malloc (1 + MY_MAX_MSG);
  if (NULL == outbuf)
    {
      perror("malloc");
      _exit(EXIT_FAILURE);
    }

  if (0 == open_ipc())
    write(pipes[1], ERRMSG1, strlen(ERRMSG1));

  while (1 == 1) 
    {
      /* block on queue and rcv the message
       */ 
      do {
	errno = 0;
	rc = msgrcv(msgid, recv_msg, MY_MAX_MSG+1, 0, 0);
      } 
      while (rc < 0 && errno == EINTR);
      
      if (rc == -1 && errno != EAGAIN && errno != ENOMSG)
	{ 
	  if (errno == EINVAL || errno == EIDRM)
	    {
	      if (0 == open_ipc())
		write(pipes[1], ERRMSG1, strlen(ERRMSG1));
	    }
	  else
	    {
	      perror("malloc");
	      _exit(EXIT_FAILURE);
	    }
	}
      else if (rc > 0)
	{
	  g_snprintf(outbuf, 1+MY_MAX_MSG, "%s\n", 
		     recv_msg->mtext);
	  outbuf[MY_MAX_MSG] = '\0'; /* pre-1.2.3 glib bug */

	  /* Ignore any errors - if nobody is listening, that's not our
	   * concern.
	   */
	  write(pipes[1], outbuf, strlen(outbuf));
	}
    }

  /* we should never reach this, as the loop above can only be terminated
   * by a signal
   */
  if (msgid != (-1))
    status = msgctl (msgid, IPC_RMID, NULL);

  _exit(EXIT_FAILURE);
}
/*
 *
 *  END CHILD PROCESS
 *
 **********************************************************************/



static time_t now_time, last_time;

static GtkWidget   * applet = NULL;
static gint     timeout_val;
static gboolean popup_val;
static gboolean makebeep_val;
static GtkWidget *propbox = NULL;   /* preferences window         */
static gint     tmp_popup_val;
static gboolean tmp_timeout_val;
static gboolean tmp_makebeep_val;

static gint
applet_save_session(GtkWidget *w,
                    const char *privcfgpath,
                    const char *globcfgpath)
{
  gnome_config_push_prefix(privcfgpath);
  gnome_config_set_int  ("options/timeout", timeout_val);
  gnome_config_set_bool ("options/popup",   popup_val);
  gnome_config_set_bool ("options/makebeep",makebeep_val);
  gnome_config_pop_prefix();

  gnome_config_sync();
  gnome_config_drop_all();

  return FALSE;
}


static void makebeep_cb( GtkWidget *widget, gpointer data)
{
  tmp_makebeep_val = GTK_TOGGLE_BUTTON(widget)->active;
  gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
  return;
}

static void popup_cb( GtkWidget *widget, gpointer data)
{
  tmp_popup_val = GTK_TOGGLE_BUTTON(widget)->active;
  gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
  return;
}

static void timeout_cb( GtkWidget *widget, gpointer data)
{
  tmp_timeout_val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
  gnome_property_box_changed(GNOME_PROPERTY_BOX(propbox));
  return;
}

static void destroy_cb( GtkWidget   *w, gpointer data)
{
  gtk_widget_destroy (propbox);
  propbox = NULL;
  return;
}

static void help_cb( GtkWidget   *w, gpointer data)
{
  GtkWidget *woosh;
  woosh = gnome_message_box_new(_("Set your preferences"), 
				 GNOME_MESSAGE_BOX_INFO,  
				 GNOME_STOCK_BUTTON_OK, NULL); 
  gnome_dialog_set_close(GNOME_DIALOG(woosh), TRUE);
  gtk_window_set_modal(GTK_WINDOW(woosh), TRUE);
  gtk_widget_show(woosh);
  return;
}

static void apply_cb( GtkWidget   *w, gint pagenum, gpointer data)
{
  timeout_val  = tmp_timeout_val;
  popup_val    = tmp_popup_val;
  makebeep_val = tmp_makebeep_val;

  gnome_config_push_prefix(APPLET_WIDGET(applet)->privcfgpath);
  gnome_config_set_int  ("options/timeout",  timeout_val);
  gnome_config_set_bool ("options/popup",    popup_val);
  gnome_config_set_bool ("options/makebeep", makebeep_val);
  gnome_config_pop_prefix();

  gnome_config_sync();
  gnome_config_drop_all();

  return;
}

void show_properties_dialog() {
  GtkWidget *cpage;
  GtkWidget *hbox;
  GtkWidget *chkbox;
  GtkWidget *label;
  GtkObject     * adj;
  GtkWidget     * spb;
 
  if(propbox) 
    return;

  tmp_popup_val    = popup_val;
  tmp_timeout_val  = timeout_val;
  tmp_makebeep_val = makebeep_val;


  propbox = gnome_property_box_new();
  gtk_window_set_modal(GTK_WINDOW(propbox), TRUE);
  gtk_window_set_title(GTK_WINDOW(&GNOME_PROPERTY_BOX(propbox)->dialog.window),
		       _("Preferences"));
  
  cpage = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(cpage), GNOME_PAD_SMALL);
  
  hbox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(cpage), hbox, FALSE, TRUE, GNOME_PAD_SMALL);
  chkbox = gtk_check_button_new_with_label(_("Enable popup warnings"));
  GTK_TOGGLE_BUTTON(chkbox)->active = popup_val;
  gtk_signal_connect(GTK_OBJECT(chkbox), "clicked",
                     (GtkSignalFunc) popup_cb, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), chkbox, TRUE, TRUE, GNOME_PAD_SMALL);
  gtk_widget_show (chkbox);
  gtk_widget_show(hbox);

  hbox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(cpage), hbox, FALSE, TRUE, GNOME_PAD_SMALL);
  chkbox = gtk_check_button_new_with_label(_("Enable bell warnings"));
  GTK_TOGGLE_BUTTON(chkbox)->active = makebeep_val;
  gtk_signal_connect(GTK_OBJECT(chkbox), "clicked",
                     (GtkSignalFunc) makebeep_cb, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), chkbox, TRUE, TRUE, GNOME_PAD_SMALL);
  gtk_widget_show (chkbox);
  gtk_widget_show(hbox);

  hbox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(cpage), hbox, FALSE, TRUE, GNOME_PAD_SMALL);
  adj = gtk_adjustment_new( (gfloat) timeout_val, 
			    10.0, 86400.0, 10.0, 100.0, 86400.0);
  spb = gtk_spin_button_new( GTK_ADJUSTMENT(adj), 0.5, 1);
  label = gtk_label_new(_("Timeout (seconds)"));
  gtk_signal_connect(GTK_OBJECT(spb), "changed",
                     (GtkSignalFunc) timeout_cb, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), spb, TRUE, TRUE, GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, GNOME_PAD_SMALL);
  gtk_widget_show(label);
  gtk_widget_show(spb);
  gtk_widget_show(hbox);

  gtk_widget_show(cpage);
  label = gtk_label_new(_("Configuration"));
  gtk_notebook_append_page(GTK_NOTEBOOK(GNOME_PROPERTY_BOX(propbox)->notebook),
                           cpage, label);

  gtk_signal_connect (GTK_OBJECT (propbox), "destroy",
                      GTK_SIGNAL_FUNC (destroy_cb), NULL);
  gtk_signal_connect (GTK_OBJECT (propbox), "help",
                      GTK_SIGNAL_FUNC (help_cb), NULL);
  gtk_signal_connect (GTK_OBJECT (propbox), "apply",
                      GTK_SIGNAL_FUNC (apply_cb), NULL);
  gtk_widget_show(propbox);
}


/**********************************************************************
 *
 *  CALLBACK ROUTINE
 *
 *  read messages from the pipe, and act upon them
 */
gboolean pipe_reader (GIOChannel *source, 
		      GIOCondition condition,  gpointer data)
{
  char     outbuf[1 + MY_MAX_MSG];
  int      fd;
  char    *res;
  static   FILE   * infile = NULL;

  data = NULL; /* ugly kludge to avoid 'unused param' warning */

  if (((condition & G_IO_IN) != 0) || ((condition & G_IO_PRI) != 0))
    {
      if (infile == NULL)
	{
	  /* We open this stream only once, and never close it
	   * (the kernel will clean up on exit)
	   */
	  fd = g_io_channel_unix_get_fd(source);
	  infile = fdopen(fd, "r");
	}

      do {
	res = fgets(outbuf, 1+MY_MAX_MSG, infile);
	if (res != NULL)
	  {
	    insert_in_buffer (outbuf);
	    last_time = time(NULL);
	  }
      } while (res != NULL && *res != '\0');

      return TRUE;
    }
  else
    {
      if      ((condition & G_IO_ERR) != 0)
	g_warning("pipe_reader: error on message pipe\n");
      else if ((condition & G_IO_HUP) != 0)
	g_warning("pipe_reader: hung up on message pipe\n");
      else if ((condition & G_IO_NVAL) != 0)
	g_warning("pipe_reader: invalid request on message pipe\n");
      applet_widget_gtk_main_quit();
    }
  return FALSE;
}

GtkWidget * label_normal;
GtkWidget * label_warn;


/* the about... callback
 */
void about_cb ()
{
  GtkWidget *about;
  const gchar *author[]={"Rainer Wichmann <beltane@la-samhna.de>",
                         NULL};
  about = gnome_about_new (PACKAGE, VERSION,
			   _("(c) 2001 Rainer Wichmann"),
			   author,
			   _("Gnome panel applet for monitoring the "\
			     "samhain file integrity checking / intrusion "
			     "detection daemon. "\
			     "Released under the "\
			     "GNU General Public License."),
			   NULL);
  gtk_widget_show(about);
  return;
}

/* Convenience function to create a pixmap widget from data
 *
 * David Bryant <d_bryant@lincoln.college.adelaide.edu.au>
 *
 * This function is taken from the 'geg' program,
 * version 0.15.0, without any modifications except for
 * error handling. Copyright is
 * with the author of 'geg', David Bryant. 
 * ('geg' is distributed under the 
 * terms of the GNU General Public License). 
 */
GtkWidget *
create_pixmap(GtkWidget *widget, gchar **data)
{
  GtkStyle *style;
  GdkBitmap *mask;
  GdkPixmap *gdk_pixmap;
  GtkWidget *gtk_pixmap;
  
  style = gtk_widget_get_style(widget);
  if (style == NULL) 
    GDIE("gtk_widget_get_style");

  gdk_pixmap = 
    gdk_pixmap_create_from_xpm_d(widget->window, 
				 &mask, &style->bg[GTK_STATE_NORMAL],
				 data);
  if (gdk_pixmap == NULL) 
    GDIE("gdk_pixmap_create_from_xpm_d");
    
  gtk_pixmap = gtk_pixmap_new(gdk_pixmap, mask);
  if (gtk_pixmap == NULL) 
    GDIE("gtk_pixmap_new");

  gtk_widget_show(gtk_pixmap);
  return(gtk_pixmap);
}

static gboolean state = FALSE;  

void switch_to_normal ()
{
  if (state == FALSE)
    return;
  gtk_widget_hide(GTK_WIDGET(label_warn));
  gtk_widget_show(GTK_WIDGET(label_normal));
  state = FALSE;
  return;
}

static void popup_warn(char * data)
{
  GtkWidget *woosh;
  char blurb[256];

  if (popup_val == FALSE)
    return;

  g_snprintf(blurb, 256, "%s%s", _("SAMHAIN DAEMON STATUS\n\n"), data);
  blurb[255] = '\0';  /* glib < 1.2.3 bug */

  woosh = gnome_message_box_new(blurb, 
				GNOME_MESSAGE_BOX_ERROR,  
				GNOME_STOCK_BUTTON_OK, 
				NULL); 
  gnome_dialog_set_close(GNOME_DIALOG(woosh), TRUE);
  gtk_window_set_modal(GTK_WINDOW(woosh), TRUE);
  gtk_widget_show(woosh);
  return;
}

void beep_warn ()
{
  if (makebeep_val == FALSE)
    return;

  gdk_beep();

  return;
}

void switch_to_warn (char * msg, char * details)
{
  char full[256];
  if (state == TRUE)
    return;
  gtk_widget_hide(GTK_WIDGET(label_normal));
  gtk_widget_show(GTK_WIDGET(label_warn));
  if (details == NULL)
    popup_warn(msg);
  else
    {
      g_snprintf(full, 256, "%s\n\n%s", msg, details);
      full[255] = '\0'; /* glib < 1.2.3 bug */
      popup_warn(full);
    }
  beep_warn();
  state = TRUE;
  return;
}

gint watchdog (gpointer data )
{
  time_t now = time(NULL);

  if ((now - last_time) > timeout_val)
    {
      switch_to_warn (_("Timeout (daemon exited or not started ?)"), NULL);
      last_time = now;
    }
  return TRUE;
}

/* Fork off the msg_handle() process to read the SysV message queue,
 * then initialize the applet, and poll the pipe for messages
 */
int main(int argc, char **argv)
{
  GtkWidget      * table;
  GtkWidget      * event_box;
  GIOChannel     * watchme;

  int              rc       = -1;
  int              pipes[2];
  int              child_pid;
  int              fdflags;

  gint             size;

  /* create the pipe
   */
  rc = pipe(pipes);
  if (rc == -1)
    {
      perror("pipe");
      exit(EXIT_FAILURE);
    }


  /* fork, close the unused end of the pipe, and set the pipe to
   * non-blocking (otherwise the fgets() below will block)
   */
  child_pid = fork();

  switch (child_pid) 
    {
    case -1:                /* fork() failed  */
      perror("fork");
      exit(1);

    case 0:                 /* child          */
      close(pipes[0]);
      fdflags = fcntl(pipes[1], F_GETFL);
      fcntl(pipes[1], F_SETFL, (long) (fdflags | O_NONBLOCK));
      msg_handle(pipes);
      
    default:                /* parent         */
      break;
    }

  /* if we have superuser privileges, we drop them here
   */
  if (0 != setuid(getuid()))
    {
      perror("setuid");
      exit(EXIT_FAILURE);
    }

  /* close unused end of pipe, set to non-blocking
   */
  close(pipes[1]);
  fdflags = fcntl(pipes[0], F_GETFL);
  fcntl(pipes[0], F_SETFL, (long) (fdflags | O_NONBLOCK));

  now_time = time(&last_time);

  /* set up signal handler for SIGCHLD and for killing the child on exit
   */
  the_child_pid = child_pid;
  init_sigchld();

  /* Initialize the i18n stuff 
   * Need the #ifdef, because the dummy supplied by gnome-i18n.h
   * will cause a compiler warning
   */
#ifdef ENABLE_NLS
  bindtextdomain (PACKAGE, GNOMELOCALEDIR);
  textdomain (PACKAGE);
#endif
  
  /* Intialize, this will basically set up the applet, corba and
   * call gnome_init 
   */
  applet_widget_init(PACKAGE, VERSION, argc, argv, NULL,0,NULL);
  
  /* create a new applet_widget 
   */
  applet = applet_widget_new("samhain_monitor_applet");

  /* in the rare case that the communication with the panel
   * failed, error out 
   */
  if (!applet)
    GDIE("applet_widget_new");
  gtk_widget_show (applet);
  
  gnome_config_push_prefix(APPLET_WIDGET(applet)->privcfgpath);
  timeout_val  = gnome_config_get_int  ("options/timeout=1200");
  popup_val    = gnome_config_get_bool ("options/popup=true");
  makebeep_val = gnome_config_get_bool ("options/makebeep=true");
  gnome_config_pop_prefix();

  /* create the widget we are going to put on the applet 
   */
  size = applet_widget_get_panel_pixel_size(APPLET_WIDGET (applet));

  table = gtk_table_new (1, 1, TRUE);

  label_normal = create_pixmap(GTK_WIDGET(applet), samhain_xpm);
  label_warn   = create_pixmap(GTK_WIDGET(applet), samhain_warn_xpm);
  gtk_table_attach_defaults (GTK_TABLE(table), label_normal, 0, 1, 0, 1);
  gtk_table_attach_defaults (GTK_TABLE(table), label_warn,   0, 1, 0, 1);

  event_box = gtk_event_box_new ();
  gtk_container_add (GTK_CONTAINER(event_box), table);
  gtk_widget_show (event_box);
  gtk_widget_show (table);

  gtk_signal_connect (GTK_OBJECT (event_box), "button_press_event",
                      (GtkSignalFunc) log_event,
                      NULL);
  gtk_widget_show(label_normal);

  /* bind the session save signal 
   */
  gtk_signal_connect(GTK_OBJECT(applet),"save_session",
		     GTK_SIGNAL_FUNC(applet_save_session),
		     NULL);
  
  /* add the widget to the applet-widget, and thereby actually
   * putting it "onto" the panel 
   */
  applet_widget_add (APPLET_WIDGET (applet), event_box);
  
  /* add the reset callback to the applet menu 
   */
  applet_widget_register_callback(APPLET_WIDGET(applet),
					"reset",
					_("Reset"),
					switch_to_normal,
					NULL);

  applet_widget_register_callback(APPLET_WIDGET(applet),
					"eventlog",
					_("Show log"),
					log_event,
					NULL);

  applet_widget_register_stock_callback(APPLET_WIDGET(applet),
					"props",
					GNOME_STOCK_MENU_PROP,
					_("Properties..."),
					show_properties_dialog,
					NULL);

  applet_widget_register_stock_callback(APPLET_WIDGET(applet),
					"about",
					GNOME_STOCK_MENU_ABOUT,
					_("About..."),
					about_cb,
					NULL);

  /* prepare for polling the pipe
   */
  watchme = g_io_channel_unix_new(pipes[0]);

  g_io_add_watch (watchme,                                     /* channel   */
		  G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,/* condition */
		  pipe_reader,                                 /* callback  */
		  NULL);                                       /* user data */

  gtk_timeout_add( /* milliseconds */ 5000, watchdog, NULL );

  /* special corba main loop 
   */
  applet_widget_gtk_main ();

  /* if the applet is removed by using the panel functions, the child
   * will survive, thus we have to kill it explicitely when leaving
   * the main loop
   */
  kill(child_pid, SIGTERM);

  return 0;
}
